@@ -21,13 +21,13 @@
 #include <string.h>
 #include <stdlib.h>
 #include <config.h>
+#include <gmodule.h>
 
 #include "playlist-treeview.h"
 #include "renderer-combo.h"
 #include "renderer-controls.h"
 #include "icons.h"
 #include "gui.h"
-#include "main.h"
 
 #define CONTENT_DIR "urn:schemas-upnp-org:service:ContentDirectory"
 
@@ -43,6 +43,9 @@ typedef gboolean (* RowCompareFunc) (GtkTreeModel *model,
                                      gpointer      user_data);
 
 static GtkWidget *treeview;
+static GtkWidget *popup;
+static GtkWidget *didl_dialog;
+static GtkWidget *didl_textview;
 static gboolean   expanded;
 
 typedef struct
@@ -56,7 +59,7 @@ typedef struct
 
 typedef struct
 {
-  GetSelectedItemCallback callback;
+  MetadataFunc callback;
 
   gchar *id;
 
@@ -87,9 +90,9 @@ browse_data_free (BrowseData *data)
 }
 
 static BrowseMetadataData *
-browse_metadata_data_new (GetSelectedItemCallback callback,
-                          const char             *id,
-                          gpointer                user_data)
+browse_metadata_data_new (MetadataFunc callback,
+                          const char  *id,
+                          gpointer     user_data)
 {
         BrowseMetadataData *data;
 
@@ -114,12 +117,36 @@ browse (GUPnPServiceProxy *content_dir,
         guint32            starting_index,
         guint32            requested_count);
 
+G_MODULE_EXPORT
 gboolean
 on_playlist_treeview_button_release (GtkWidget      *widget,
                                      GdkEventButton *event,
                                      gpointer        user_data)
 {
-        return FALSE;
+        GtkTreeSelection *selection;
+        GtkTreeModel     *model;
+        GtkTreeIter       iter;
+
+        if (event->type != GDK_BUTTON_RELEASE ||
+            event->button != 3)
+                return FALSE;
+
+        selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (treeview));
+        g_assert (selection != NULL);
+
+        /* Only show the popup menu when a row is selected */
+        if (!gtk_tree_selection_get_selected (selection, &model, &iter))
+                return FALSE;
+
+        gtk_menu_popup (GTK_MENU (popup),
+                        NULL,
+                        NULL,
+                        NULL,
+                        NULL,
+                        event->button,
+                        event->time);
+
+        return TRUE;
 }
 
 static void
@@ -132,9 +159,8 @@ on_item_selected (GtkTreeSelection *selection,
 
         if (state == PLAYBACK_STATE_PLAYING ||
             state == PLAYBACK_STATE_PAUSED) {
-                if (!get_selected_item ((GetSelectedItemCallback)
-                                         set_av_transport_uri,
-                                        NULL)) {
+                if (!get_selected_object ((MetadataFunc) set_av_transport_uri,
+                                          NULL)) {
                         av_transport_send_action ("Stop", NULL);
                 }
        }
@@ -159,7 +185,7 @@ create_playlist_treemodel (void)
                                     /* Is container? */
                                     G_TYPE_BOOLEAN,
                                     /* childCount */
-                                    G_TYPE_UINT);
+                                    G_TYPE_INT);
 
         return GTK_TREE_MODEL (store);
 }
@@ -200,6 +226,7 @@ setup_treeview_columns (GtkWidget *treeview)
         gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);
 }
 
+G_MODULE_EXPORT
 void
 on_playlist_row_expanded (GtkTreeView *tree_view,
                           GtkTreeIter *iter,
@@ -220,7 +247,7 @@ on_playlist_row_expanded (GtkTreeView *tree_view,
                 GUPnPServiceProxy *content_dir;
                 gchar             *id;
                 gboolean           is_container;
-                guint              child_count;
+                gint              child_count;
 
                 gtk_tree_model_get (model, &child_iter,
                                     3, &content_dir,
@@ -229,7 +256,7 @@ on_playlist_row_expanded (GtkTreeView *tree_view,
                                     6, &child_count,
                                     -1);
 
-                if (is_container && child_count > 0) {
+                if (is_container && child_count != 0) {
                         browse (content_dir, id, 0, MAX_BROWSE);
                 }
 
@@ -254,6 +281,7 @@ unpopulate_container (GtkTreeModel *model,
                                      &child_iter));
 }
 
+G_MODULE_EXPORT
 void
 on_playlist_row_collapsed (GtkTreeView *tree_view,
                            GtkTreeIter *iter,
@@ -275,31 +303,23 @@ on_playlist_row_collapsed (GtkTreeView *tree_view,
         } while (gtk_tree_model_iter_next (model, &child_iter));
 }
 
-static gboolean
-tree_selection_func (GtkTreeSelection *selection,
-                     GtkTreeModel     *model,
-                     GtkTreePath      *path,
-                     gboolean          path_currently_selected,
-                     gpointer          data)
+static void display_metadata (const char *metadata,
+                              gpointer    user_data)
 {
-        GtkTreeIter iter;
-        gboolean    is_container;
+        GtkTextBuffer *buffer;
 
-        if (path_currently_selected) {
-                return TRUE;
-        }
+        buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (didl_textview));
+        gtk_text_buffer_set_text (buffer, metadata, -1);
 
-        if (!gtk_tree_model_get_iter (model, &iter, path)) {
-                return FALSE;
-        }
+        gtk_dialog_run (GTK_DIALOG (didl_dialog));
+}
 
-        gtk_tree_model_get (model,
-                            &iter,
-                            5, &is_container,
-                            -1);
-
-        /* Let it be selected if it's an item */
-        return !is_container;
+G_MODULE_EXPORT
+void
+on_didl_menuitem_activate (GtkMenuItem *menuitem,
+                           gpointer     user_data)
+{
+        get_selected_object (display_metadata, NULL);
 }
 
 void
@@ -312,6 +332,14 @@ setup_playlist_treeview (GtkBuilder *builder)
                                                        "playlist-treeview"));
         g_assert (treeview != NULL);
 
+        popup = GTK_WIDGET (gtk_builder_get_object (builder, "playlist-popup"));
+        g_assert (popup != NULL);
+
+        didl_dialog = GTK_WIDGET (gtk_builder_get_object (builder,
+                                                          "didl-dialog"));
+        didl_textview = GTK_WIDGET (gtk_builder_get_object (builder,
+                                                          "didl-textview"));
+
         model = create_playlist_treemodel ();
         g_assert (model != NULL);
 
@@ -326,10 +354,6 @@ setup_playlist_treeview (GtkBuilder *builder)
                           "changed",
                           G_CALLBACK (on_item_selected),
                           NULL);
-        gtk_tree_selection_set_select_function (selection,
-                                                tree_selection_func,
-                                                NULL,
-                                                NULL);
 
         expanded = FALSE;
 }
@@ -584,7 +608,7 @@ append_didl_object (GUPnPDIDLLiteObject *object,
         const char *parent_id;
         const char *title;
         gboolean    is_container;
-        uint        child_count;
+        gint        child_count;
         GdkPixbuf  *icon;
         gint        position;
 
@@ -751,7 +775,7 @@ browse_cb (GUPnPServiceProxy       *content_dir,
                 if (!gupnp_didl_lite_parser_parse_didl (parser,
                                                         didl_xml,
                                                         &error)) {
-                        g_warning ("%s\n", error->message);
+                        g_warning ("%s", error->message);
                         g_error_free (error);
                 }
 
@@ -774,7 +798,7 @@ browse_cb (GUPnPServiceProxy       *content_dir,
                 GUPnPServiceInfo *info;
 
                 info = GUPNP_SERVICE_INFO (content_dir);
-                g_warning ("Failed to browse '%s': %s\n",
+                g_warning ("Failed to browse '%s': %s",
                            gupnp_service_info_get_location (info),
                            error->message);
 
@@ -784,6 +808,44 @@ browse_cb (GUPnPServiceProxy       *content_dir,
         browse_data_free (data);
 }
 
+static void
+browse_metadata_cb (GUPnPServiceProxy       *content_dir,
+                    GUPnPServiceProxyAction *action,
+                    gpointer                 user_data)
+{
+        BrowseMetadataData *data;
+        char               *metadata;
+        GError             *error;
+
+        data = (BrowseMetadataData *) user_data;
+
+        metadata = NULL;
+        error = NULL;
+
+        gupnp_service_proxy_end_action (content_dir,
+                                        action,
+                                        &error,
+                                        /* OUT args */
+                                        "Result",
+                                        G_TYPE_STRING,
+                                        &metadata,
+                                        NULL);
+        if (metadata) {
+                data->callback (metadata, data->user_data);
+
+                g_free (metadata);
+        } else if (error) {
+                g_warning ("Failed to get metadata for '%s': %s",
+                           data->id,
+                           error->message);
+
+                g_error_free (error);
+        }
+
+        browse_metadata_data_free (data);
+        g_object_unref (content_dir);
+}
+
 static void
 browse (GUPnPServiceProxy *content_dir,
         const char        *container_id,
@@ -823,6 +885,42 @@ browse (GUPnPServiceProxy *content_dir,
                                  NULL);
 }
 
+static void
+browse_metadata (GUPnPServiceProxy *content_dir,
+                 const char        *id,
+                 MetadataFunc       callback,
+                 gpointer           user_data)
+{
+        BrowseMetadataData *data;
+
+        data = browse_metadata_data_new (callback, id, user_data);
+
+        gupnp_service_proxy_begin_action
+                                (g_object_ref (content_dir),
+                                 "Browse",
+                                 browse_metadata_cb,
+                                 data,
+                                 /* IN args */
+                                 "ObjectID",
+                                 G_TYPE_STRING,
+                                 data->id,
+                                 "BrowseFlag",
+                                 G_TYPE_STRING,
+                                 "BrowseMetadata",
+                                 "Filter",
+                                 G_TYPE_STRING,
+                                 "*",
+                                 "StartingIndex",
+                                 G_TYPE_UINT,
+                                 0,
+                                 "RequestedCount",
+                                 G_TYPE_UINT, 0,
+                                 "SortCriteria",
+                                 G_TYPE_STRING,
+                                 "",
+                                 NULL);
+}
+
 static void
 append_media_server (GUPnPDeviceProxy *proxy,
                      GtkTreeModel     *model,
@@ -848,6 +946,7 @@ append_media_server (GUPnPDeviceProxy *proxy,
                                  1, friendly_name,
                                  2, info,
                                  3, content_dir,
+                                 4, "0",
                                  5, TRUE,
                                  -1);
                 g_free (friendly_name);
@@ -930,54 +1029,14 @@ remove_media_server (GUPnPDeviceProxy *proxy)
         }
 }
 
-void
-browse_metadata_cb (GUPnPServiceProxy       *content_dir,
-                    GUPnPServiceProxyAction *action,
-                    gpointer                 user_data)
-{
-        BrowseMetadataData *data;
-        char               *metadata;
-        GError             *error;
-
-        data = (BrowseMetadataData *) user_data;
-
-        metadata = NULL;
-        error = NULL;
-
-        gupnp_service_proxy_end_action (content_dir,
-                                        action,
-                                        &error,
-                                        /* OUT args */
-                                        "Result",
-                                        G_TYPE_STRING,
-                                        &metadata,
-                                        NULL);
-        if (metadata) {
-                data->callback (metadata, data->user_data);
-
-                g_free (metadata);
-        } else if (error) {
-                g_warning ("Failed to get metadata for '%s': %s\n",
-                           data->id,
-                           error->message);
-
-                g_error_free (error);
-        }
-
-        browse_metadata_data_free (data);
-        g_object_unref (content_dir);
-}
-
 gboolean
-get_selected_item (GetSelectedItemCallback callback,
-                   gpointer                user_data)
+get_selected_object (MetadataFunc callback,
+                     gpointer     user_data)
 {
         GUPnPServiceProxy  *content_dir;
         GtkTreeSelection   *selection;
         GtkTreeModel       *model;
         GtkTreeIter         iter;
-        gboolean            is_container;
-        BrowseMetadataData *data;
         char               *id = NULL;
         gboolean            ret = FALSE;
 
@@ -992,42 +1051,12 @@ get_selected_item (GetSelectedItemCallback callback,
                             &iter,
                             3, &content_dir,
                             4, &id,
-                            5, &is_container,
                             -1);
 
-        if (is_container) {
-                goto free_and_return;
-        }
+        browse_metadata (g_object_ref (content_dir), id, callback, user_data);
 
-        data = browse_metadata_data_new (callback, id, user_data);
-
-        gupnp_service_proxy_begin_action
-                                (g_object_ref (content_dir),
-                                 "Browse",
-                                 browse_metadata_cb,
-                                 data,
-                                 /* IN args */
-                                 "ObjectID",
-                                 G_TYPE_STRING,
-                                 data->id,
-                                 "BrowseFlag",
-                                 G_TYPE_STRING,
-                                 "BrowseMetadata",
-                                 "Filter",
-                                 G_TYPE_STRING,
-                                 "*",
-                                 "StartingIndex",
-                                 G_TYPE_UINT,
-                                 0,
-                                 "RequestedCount",
-                                 G_TYPE_UINT, 0,
-                                 "SortCriteria",
-                                 G_TYPE_STRING,
-                                 "",
-                                 NULL);
         ret = TRUE;
 
-free_and_return:
         if (id) {
                 g_free (id);
         }
